@@ -1,8 +1,9 @@
 # coding: utf8
 from __future__ import print_function
+import copy
 import logging
-import sys
 import os
+import sys
 import re
 import webbrowser
 
@@ -11,11 +12,11 @@ import wsgiref.validate
 from wsgiref.simple_server import make_server
 try:
     from http.cookies import BaseCookie
-    from urllib.parse import quote, unquote, urlparse, urlunparse
+    from urllib.parse import quote, unquote, urlparse, urlunparse, urlencode
 except ImportError:
     from Cookie import BaseCookie  # NOQA
     from urlparse import urlparse, urlunparse  # NOQA
-    from urllib import quote, unquote  # NOQA
+    from urllib import quote, unquote, urlencode  # NOQA
 
 from functools import wraps
 from itertools import chain
@@ -27,11 +28,11 @@ from lxml.cssselect import CSSSelector, SelectorSyntaxError
 from lxml.etree import XPath, XPathError
 
 from fresco import Request, Response
-from fresco.util.urls import url_join, make_query
+from fresco.util.urls import url_join
 from fresco.util.http import parse_header
 from fresco.util.wsgi import unicode_to_environ
 
-__version__ = '7.4.4'
+__version__ = '7.4.5'
 
 #: Registry for xpath multimethods
 xpath_registry = {}
@@ -43,9 +44,11 @@ REGEXP_NAMESPACE = "http://exslt.org/regular-expressions"
 if sys.version_info > (3, 0):
     PY3 = True
     string_types = str
+    ustr = str
 else:
     PY3 = False
     string_types = (unicode, str)
+    ustr = unicode
 
 
 class BadStatusError(AssertionError):
@@ -585,7 +588,7 @@ class ElementWrapper(object):
         else:
             # List of values, eg ``['1', '23', '8']``
             found = set()
-            values = set(str(v) for v in values)
+            values = set(ustr(v) for v in values)
             for el in self.input_group():
                 if el.attrib.get('value') in values:
                     el.attrib['checked'] = ""
@@ -607,7 +610,7 @@ class ElementWrapper(object):
         button in the group with the given value and checking it.
         """
         if value is not None:
-            value = str(value)
+            value = ustr(value)
         found = False
         for el in self.el.xpath(
             "./ancestor-or-self::form[1]//input[@type='radio' and @name=$n]",
@@ -630,7 +633,7 @@ class ElementWrapper(object):
         Set the value of a textarea control
         """
         if value is not None:
-            value = str(value)
+            value = ustr(value)
         self.el.text = value
         return self
 
@@ -659,7 +662,7 @@ class ElementWrapper(object):
             except KeyError:
                 pass
         else:
-            self.value = str(value)
+            self.value = ustr(value)
         return self
 
     @when("select[@multiple]")  # NOQA
@@ -677,7 +680,7 @@ class ElementWrapper(object):
             ]
 
         found = set()
-        values = set(str(v) for v in values)
+        values = set(ustr(v) for v in values)
         for opt in options:
             value = opt.attrib.get('value', opt.text)
             if value in values:
@@ -700,7 +703,7 @@ class ElementWrapper(object):
         :param values: list of values to be selected
         """
         if value is not None:
-            value = str(value)
+            value = ustr(value)
         found = False
         for opt in self.el.xpath('.//option'):
             if opt.attrib.get('value', opt.text) == value:
@@ -929,7 +932,6 @@ class TestAgent(object):
         :param response: The ``fresco.request.Response`` object associated with
                          the current request
         """
-
         if logger is not None:
             logger = logger
         elif loglevel is not None:
@@ -948,6 +950,10 @@ class TestAgent(object):
 
         self.request = request
         self.response = response
+        if request:
+            self.original_environ = request.environ.copy()
+        else:
+            self.original_environ = {}
 
         #: The original wsgi application
         self.app = app
@@ -1036,7 +1042,7 @@ class TestAgent(object):
         environ = self.environ_defaults.copy()
         environ.update(kwargs)
         for key, value in kwargs.items():
-            if '.' not in key:
+            if '.' not in key and isinstance(value, ustr):
                 value = unicode_to_environ(value)
             environ[key.replace('wsgi_', 'wsgi.')] = value
 
@@ -1101,7 +1107,8 @@ class TestAgent(object):
             if postdata:
                 self.options['logger'].info("wsgi.input: %r", postdata)
 
-        response = self.response_class.from_wsgi(self.validated_app, environ,
+        response = self.response_class.from_wsgi(self.validated_app,
+                                                 environ.copy(),
                                                  self.start_response)
         agent = self.__class__(
             self.app,
@@ -1134,7 +1141,7 @@ class TestAgent(object):
                  method='POST',
                  **kwargs):
         """
-        Make a POST request to the application and return the response.
+        Make an HTTP request to the application and return the response.
 
         :param PATH_INFO: The path to request from the application. This must
                           be URL encoded.
@@ -1174,7 +1181,7 @@ class TestAgent(object):
             envargs = kwargs
         else:
             if not isinstance(data, bytes):
-                data = make_query(data, charset=charset).encode('ASCII')
+                data = _urlencode(data, encoding=charset).encode('ASCII')
             envargs = {'CONTENT_TYPE': content_type,
                        'CONTENT_LENGTH': str(len(data))}
             envargs.update(kwargs)
@@ -1197,8 +1204,8 @@ class TestAgent(object):
     def get(self, PATH_INFO='/', data=None, charset='UTF-8', *args, **kwargs):
         kwargs.setdefault('method', 'GET')
         if data is not None:
-            kwargs.setdefault('QUERY_STRING', make_query(data,
-                                                         charset=charset))
+            kwargs.setdefault('QUERY_STRING', _urlencode(data,
+                                                         encoding=charset))
             data = None
         return self._request(PATH_INFO, data, charset, *args, **kwargs)
 
@@ -1317,7 +1324,7 @@ class TestAgent(object):
         """
         if self.options['logger']:
             self.options['logger'].debug("Reload: %s", self.request.path)
-        env = self.request.environ.copy()
+        env = self.original_environ.copy()
         del env['fresco.request']
 
         wsgi_input = env['wsgi.input']
@@ -1328,40 +1335,35 @@ class TestAgent(object):
         agent = self._wsgi_request(env, follow=follow,
                                    check_status=check_status)
 
-        for src, dst in zip(
-            self.find('//input|//textarea|//option'),
-            agent.find('//input|//textarea|//option')
-        ):
-            if not all([
-                src.tag == dst.tag,
-                src.attrib.get('name') == dst.attrib.get('name'),
-                src.attrib.get('type') == dst.attrib.get('type'),
-            ]):
-                break
+        if self._lxml is not None:
+            for src, dst in zip(self.find('//input|//textarea|//option'),
+                                agent.find('//input|//textarea|//option')):
+                if not all([src.tag == dst.tag,
+                            src.attrib.get('name') == dst.attrib.get('name'),
+                            src.attrib.get('type') == dst.attrib.get('type')]):
+                    break
 
-            if (
-                src.tag == 'input'
-                and src.attrib.get('type') in ('radio', 'checkbox')
-            ):
-                if 'checked' in src.attrib:
-                    dst.attrib['checked'] = src.attrib['checked']
-                elif 'checked' in dst.attrib:
-                    del dst.attrib['checked']
+                if src.tag == 'input' and \
+                        src.attrib.get('type') in ('radio', 'checkbox'):
+                    if 'checked' in src.attrib:
+                        dst.attrib['checked'] = src.attrib['checked']
+                    elif 'checked' in dst.attrib:
+                        del dst.attrib['checked']
 
-            elif src.tag == 'option':
-                if 'selected' in src.attrib:
-                    dst.attrib['selected'] = src.attrib['selected']
-                elif 'selected' in dst.attrib:
-                    del dst.attrib['selected']
+                elif src.tag == 'option':
+                    if 'selected' in src.attrib:
+                        dst.attrib['selected'] = src.attrib['selected']
+                    elif 'selected' in dst.attrib:
+                        del dst.attrib['selected']
 
-            elif src.tag == 'textarea':
-                dst.el.text = src.el.text
+                elif src.tag == 'textarea':
+                    dst.el.text = src.el.text
 
-            else:
-                if 'value' in src.attrib:
-                    dst.attrib['value'] = src.attrib['value']
-                elif 'value' in dst.attrib:
-                    del dst.attrib['value']
+                else:
+                    if 'value' in src.attrib:
+                        dst.attrib['value'] = src.attrib['value']
+                    elif 'value' in dst.attrib:
+                        del dst.attrib['value']
 
         return agent
 
@@ -1418,6 +1420,38 @@ class TestAgent(object):
     def root_element(self):
         return ElementWrapper(self, self.lxml)
 
+    def html(self):
+        """
+        Return an HTML representation of the (html) response's root element
+
+        :rtype: unicode string
+        """
+        return self.root_element.html()
+
+    def pretty(self):
+        """
+        Return an pretty-printed string representation of the (html) response
+        body
+
+        :rtype: unicode string
+        """
+        return self.root_element.pretty()
+
+    def striptags(self):
+        """
+        Return the (html) response's root element, with all tags stripped out,
+        leaving only the textual content. Normalizes all sequences of
+        whitespace to a single space.
+
+        Use this for simple text comparisons when testing for document content
+
+        :rtype: unicode string
+        """
+        return self.root_element.striptags()
+
+    def __contains__(self, what):
+        return what in self.body
+
     def reset(self):
         """
         Reset the lxml document, abandoning any changes made
@@ -1587,14 +1621,7 @@ class TestAgent(object):
         Return a new TestAgent with all cookies deleted.
         This gives an easy way to test session expiry.
         """
-        agent = self.__class__(
-            self.app,
-            self.request,
-            self.response,
-            None,
-            self.history,
-            **self.options
-        )
+        agent = copy.copy(self)
         agent.cookies = BaseCookie()
         return agent
 
@@ -1823,3 +1850,17 @@ def base_url(environ):
 
     url += quote(environ.get('SCRIPT_NAME', ''))
     return url
+
+
+def _urlencode_items(data, encoding):
+    return urlencode([(k if isinstance(k, bytes) else ustr(k).encode(encoding),
+                       v if isinstance(v, bytes) else ustr(v).encode(encoding))
+                       for k, v in data])
+
+
+def _urlencode(data, encoding):
+    if hasattr(data, 'allitems'):
+        return _urlencode_items(data.allitems(), encoding)
+    if hasattr(data, 'items'):
+        return _urlencode_items(data.items(), encoding)
+    return _urlencode_items(data, encoding)
